#!/usr/bin/env python

"""
Moin wiki format document tree nodes.

Copyright (C) 2017 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

class Container:

    "A container of document nodes."

    def __init__(self, nodes):
        self.nodes = nodes

    def append(self, node):
        self.nodes.append(node)

    add = append

    append_inline = append

    def empty(self):
        return not self.nodes

    def node(self, index):
        try:
            return self.nodes[index]
        except IndexError:
            return None

    def normalise(self):

        "Combine adjacent text nodes."

        nodes = self.nodes
        self.nodes = []
        text = None

        for node in nodes:

            # Open a text node or merge text into an open node.

            if isinstance(node, Text):
                if not text:
                    text = node
                else:
                    text.merge(node)

            # Close any open text node and append the current node.

            else:
                if text:
                    self.append(text)
                    text = None
                self.append(node)

        # Add any open text node.

        if text:
            self.append(text)

    def __str__(self):
        return self.prettyprint()

    def _prettyprint(self, l, indent=""):
        for node in self.nodes:
            l.append(node.prettyprint(indent + "  "))
        return "\n".join(l)

    def _to_string(self, out):
        for node in self.nodes:
            node.to_string(out)

class Region(Container):

    "A region of the page."

    transparent_region_types = ["wiki"]

    def __init__(self, nodes, level=0, indent=0, type=None):
        Container.__init__(self, nodes)
        self.level = level
        self.indent = indent
        self.type = type

    def add(self, node):
        last = self.node(-1)
        if last and last.empty():
            self.nodes[-1] = node
        else:
            self.nodes.append(node)

    def append_inline(self, s):
        if self.is_transparent():
            self.nodes[-1].append(s)
        else:
            self.append(s)

    def have_end(self, s):
        return self.level and s.startswith("}") and self.level == len(s)

    def is_transparent(self):
        return not self.level or self.type in self.transparent_region_types

    def __repr__(self):
        return "Region(%r, %r, %r, %r)" % (self.nodes, self.level, self.indent, self.type)

    def prettyprint(self, indent=""):
        l = ["%sRegion: level=%d indent=%d type=%s" % (indent, self.level, self.indent, self.type)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_region(self.level, self.indent, self.type)
        self._to_string(out)
        out.end_region(self.level, self.indent, self.type)



class Block(Container):

    "A block in the page."

    def __repr__(self):
        return "Block(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sBlock" % indent]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_block()
        self._to_string(out)
        out.end_block()

class DefItem(Container):

    "A definition item."

    def __init__(self, nodes, pad, extra):
        Container.__init__(self, nodes)
        self.pad = pad
        self.extra = extra

    def __repr__(self):
        return "DefItem(%r, %r, %r)" % (self.nodes, self.pad, self.extra)

    def prettyprint(self, indent=""):
        l = ["%sDefItem: pad=%r extra=%r" % (indent, self.pad, self.extra)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_defitem(self.pad, self.extra)
        self._to_string(out)
        out.end_defitem(self.pad, self.extra)

class DefTerm(Container):

    "A definition term."

    def __init__(self, nodes, pad):
        Container.__init__(self, nodes)
        self.pad = pad

    def __repr__(self):
        return "DefTerm(%r, %r)" % (self.nodes, self.pad)

    def prettyprint(self, indent=""):
        l = ["%sDefTerm: pad=%r" % (indent, self.pad)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_defterm(self.pad)
        self._to_string(out)
        out.end_defterm(self.pad)

class FontStyle(Container):

    "Emphasised and/or strong text."

    def __init__(self, nodes, emphasis=False, strong=False):
        Container.__init__(self, nodes)
        self.emphasis = emphasis
        self.strong = strong

    def close_emphasis(self):
        if self.strong:
            span = FontStyle(self.nodes, emphasis=True)
            self.nodes = [span]
            self.emphasis = False
        return self.strong

    def close_strong(self):
        if self.emphasis:
            span = FontStyle(self.nodes, strong=True)
            self.nodes = [span]
            self.strong = False
        return self.emphasis

    def __repr__(self):
        return "FontStyle(%r, %r, %r)" % (self.nodes, self.emphasis, self.strong)

    def prettyprint(self, indent=""):
        l = ["%sFontStyle: emphasis=%r strong=%r" % (indent, self.emphasis, self.strong)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        if self.emphasis:
            out.start_emphasis()
        elif self.strong:
            out.start_strong()
        self._to_string(out)
        if self.emphasis:
            out.end_emphasis()
        elif self.strong:
            out.end_strong()

class Heading(Container):

    "A heading."

    def __init__(self, nodes, level, start_extra="", start_pad="", end_pad="", end_extra=""):
        Container.__init__(self, nodes)
        self.level = level
        self.start_extra = start_extra
        self.start_pad = start_pad
        self.end_pad = end_pad
        self.end_extra = end_extra

    def __repr__(self):
        return "Heading(%r, %d, %r, %r, %r, %r)" % (
            self.nodes, self.level, self.start_extra, self.start_pad, self.end_pad, self.end_extra)

    def prettyprint(self, indent=""):
        l = ["%sHeading: level=%d start_extra=%r start_pad=%r end_pad=%r end_extra=%r" % (
                indent, self.level, self.start_extra, self.start_pad, self.end_pad, self.end_extra)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_heading(self.level, self.start_extra, self.start_pad)
        self._to_string(out)
        out.end_heading(self.level, self.end_pad, self.end_extra)

class ListItem(Container):

    "A list item."

    def __init__(self, nodes, indent, marker, space):
        Container.__init__(self, nodes)
        self.indent = indent
        self.marker = marker
        self.space = space

    def __repr__(self):
        return "ListItem(%r, %r, %r, %r)" % (self.nodes, self.indent, self.marker, self.space)

    def prettyprint(self, indent=""):
        l = ["%sListItem: indent=%d marker=%r space=%r" % (indent, self.indent, self.marker, self.space)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_listitem(self.indent, self.marker, self.space)
        self._to_string(out)
        out.end_listitem(self.indent, self.marker)

class Monospace(Container):

    "Monospaced text."

    def __repr__(self):
        return "Monospace(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sMonospace" % indent]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_monospace()
        self._to_string(out)
        out.end_monospace()

class Subscript(Container):

    "Subscripted text."

    def __repr__(self):
        return "Subscript(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sSubscript" % indent]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_subscript()
        self._to_string(out)
        out.end_subscript()

class Superscript(Container):

    "Superscripted text."

    def __repr__(self):
        return "Superscript(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sSuperscript" % indent]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_superscript()
        self._to_string(out)
        out.end_superscript()

class Underline(Container):

    "Underlined text."

    def __repr__(self):
        return "Underline(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sUnderline" % indent]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_underline()
        self._to_string(out)
        out.end_underline()



class Node:

    "A document node without children."

    def empty(self):
        return False

class Break(Node):

    "A paragraph break."

    def __repr__(self):
        return "Break()"

    def prettyprint(self, indent=""):
        return "%sBreak" % indent

    def to_string(self, out):
        out.break_()

class Rule(Node):

    "A horizontal rule."

    def __init__(self, length):
        self.length = length

    def __repr__(self):
        return "Rule(%d)" % self.length

    def prettyprint(self, indent=""):
        return "%sRule: %d" % (indent, self.length)

    def to_string(self, out):
        out.rule(self.length)

class Text(Node):

    "A text node."

    def __init__(self, s):
        self.s = s

    def empty(self):
        return not self.s

    def merge(self, text):
        self.s += text.s

    def __repr__(self):
        return "Text(%r)" % self.s

    def prettyprint(self, indent=""):
        return "%sText: %r" % (indent, self.s)

    def to_string(self, out):
        out.text(self.s)

# vim: tabstop=4 expandtab shiftwidth=4
